/*
     This file is part of GNUnet.
     (C) 2001-2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * ContentDatabase (mysql implementation).
 *
 * This is not for the faint hearted and requires manual setup
 * from the MySQL side. Its also still quite beta. I just got fed up 
 * with both tdb and gdbm messing things up all the time and not scaling up, 
 * and decided to really find out how a nasty big dbmgr performs in this 
 * situation.
 *
 * How to use: In gnunet.conf, set
 *
 *  DATABASETYPE = mysql
 *  DATABASEBUCKETS = 1 
 *  
 * With mysql, do the following.
 *
    CREATE DATABASE gnunet;
    USE gnunet;
    CREATE TABLE data (
      content blob NOT NULL,
      hash varchar(40) NOT NULL default '',
      priority int(11) NOT NULL default '0',
      PRIMARY KEY  (hash),
      KEY priority (priority)
    ) TYPE=MyISAM;
 *	  
 * (This DB definition, at this point, might change without a notice,
 *  maybe making your old DB incompatible).
 *
 * Make sure gnunet-running user has read/write access to mysql
 * database "gnunet" (see MySQL manual how to grant accesses). 
 *
 * Thats all.
 *
 * TODO: No disk quotas are enforced or kept up yet
 * 
 * @author Igor Wronsky
 * @file applications/afs/database/contentdatabase_mysql.c
 **/

#include "config.h"
#include "applications/afs/database/contentdatabase_mysql.h"
#include "util/xmalloc.h"
#include "util/logging.h"
#include "util/semaphore.h"
#include "util/storage.h"
#include "util/returnvalues.h"
#include "util/statuscalls.h"

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <dirent.h>
#include <errno.h>

void * libInitContentDatabase(char * dir) {
  mysqlHandle * dbh;
  
  dbh = MALLOC(sizeof(mysqlHandle));
  dbh->dbf=mysql_init(NULL);
  if(dbh->dbf == NULL) {
    LOG(LOG_ERROR, 
	"ERROR: Unable to init MySQL\n");
    FREE(dbh);
    return NULL;
  }

  mysql_options(dbh->dbf, MYSQL_READ_DEFAULT_GROUP, "client");
  mysql_real_connect(dbh->dbf,NULL,NULL,NULL,"gnunet",0,NULL,0);
  if(mysql_error(dbh->dbf)[0]) {
    LOG(LOG_ERROR, 
	"ERROR: %s\n", mysql_error(dbh->dbf));
    FREE(dbh);
    return NULL;
  }
    
  MUTEX_CREATE_RECURSIVE(&dbh->DATABASE_Lock_);	

  return dbh;
}

/**
 * Clean shutdown of the storage module (not used at the moment)
 **/
void libDoneContentDatabase(mysqlHandle * dbh) {
 
  MUTEX_DESTROY(&dbh->DATABASE_Lock_);
  mysql_close(dbh->dbf);
  FREE(dbh);
}


/**
 * Call a method for each key in the database and
 * call the callback method on it. 
 *
 * @param callback the callback method
 * @param data second argument to all callback calls
 * @return the number of items stored in the content database
 **/
int libForEachEntryInDatabase(mysqlHandle * dbh,
 		           EntryCallback callback,
	    	           void * data) {
  MYSQL_RES *sql_res;
  MYSQL_ROW sql_row; 
  HashCode160 doubleHash;
  int count = 0;

  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  mysql_query(dbh->dbf, 
	      "SELECT hash FROM data");
  if(mysql_error(dbh->dbf)[0]) {
    LOG(LOG_ERROR, 
	"ERROR: %s\n", mysql_error(dbh->dbf));
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return(SYSERR);
  }
  
  if(!(sql_res=mysql_use_result(dbh->dbf))) {
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return(SYSERR);
  }

  while((sql_row=mysql_fetch_row(sql_res))) {
    hex2hash((HexName*)sql_row[0], 
             &doubleHash);
    callback(&doubleHash, data);
    count++;
  }
		
  mysql_free_result(sql_res);
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  return count;
}

/**
 * Get the number of entries in the database.
 **/
int libCountContentEntries(mysqlHandle * dbh) {
  MYSQL_RES * sql_res;
  MYSQL_ROW sql_row;
  int count=0;

  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  mysql_query(dbh->dbf,
  	      "SELECT count(*) FROM data");
  if(!(sql_res=mysql_store_result(dbh->dbf))) {
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return(SYSERR);
  }

  if((sql_row=mysql_fetch_row(sql_res))) {
    count = atol(sql_row[0]);
  }
  mysql_free_result(sql_res);
     
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  return count;
}

/**
 * Read the contents of a bucket to a buffer.
 *
 * @param fn the hashcode representing the entry
 * @param result the buffer to write the result to 
 *        (*result should be NULL, sufficient space is allocated)
 * @return the number of bytes read on success, -1 on failure
 **/ 
int libReadContent(mysqlHandle * dbh,
	        HashCode160 * name,
	        void ** result) {
  MYSQL_RES * sql_res;
  MYSQL_ROW sql_row;
  HexName fn;
  char * scratch;
  int len;

  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  hash2hex(name, &fn);
  scratch = MALLOC(sizeof(HexName)*2+100+1);
  sprintf(scratch, 
	  "SELECT content FROM data WHERE hash='%s'",
          fn.data);       
  mysql_query(dbh->dbf, scratch);
//  LOG(LOG_WARNING, "QUE2: %s\n", scratch);
  FREE(scratch);
  if(mysql_error(dbh->dbf)[0]) {
    LOG(LOG_ERROR, 
	"ERROR2: %s\n", mysql_error(dbh->dbf));
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return SYSERR;
  }

  if(!(sql_res=mysql_store_result(dbh->dbf))) {
    LOG(LOG_ERROR, 
	"ERROR3: %s\n", mysql_error(dbh->dbf));
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return SYSERR;
  }

  if(!(sql_row=mysql_fetch_row(sql_res))) {
    /* not error, just data not found */
    mysql_free_result(sql_res);
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return SYSERR;
  }
    
  len=mysql_fetch_lengths(sql_res)[0];
  *result = MALLOC(len);
  memcpy(*result,
         sql_row[0],
	 len);
  mysql_free_result(sql_res);
  
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  return len;
}

/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 *
 * @param fn the key for the entry
 * @param len the size of the block
 * @param block the data to store
 * @return SYSERR on error, OK if ok.
 **/
int libWriteContent(mysqlHandle * dbh,
		 HashCode160 * name, 
		 int len,
		 void * block) {
  HexName fn;
  char * convblock;
  char * scratch;
 
  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  hash2hex(name, &fn);
  convblock = MALLOC(2*len+1); 
  mysql_escape_string(convblock, (char *)block, len);
  scratch = MALLOC(len*2+sizeof(HexName)*2+100+1);
  sprintf(scratch, 
	  "REPLACE INTO data (content,hash) VALUES ('%s','%s')",
  	  convblock, fn.data);
//  LOG(LOG_WARNING, "QUE: %s\n", scratch);
  mysql_query(dbh->dbf, scratch);
  FREE(scratch);
  FREE(convblock);
  if(mysql_error(dbh->dbf)[0]) {
    LOG(LOG_ERROR, 
	"ERROR1: %s\n", mysql_error(dbh->dbf));
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return SYSERR;
  }
    
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  return OK;
}

/**
 * Free space in the database by removing one file
 * @param name hashcode representing the name of the file (without directory)
 **/
int libUnlinkFromDB(mysqlHandle * dbh,
		 HashCode160 * name) {
  HexName fn;
  char * scratch;

  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  hash2hex(name, &fn);
  scratch=MALLOC(sizeof(HexName)*2+100+1);
  sprintf(scratch, 
	  "DELETE FROM data WHERE hash='%s'",	
  	fn.data);
  mysql_query(dbh->dbf,scratch);
  FREE(scratch);
  if(mysql_error(dbh->dbf)[0]) {
    LOG(LOG_ERROR, 
	"ERROR: %s\n", mysql_error(dbh->dbf));
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return SYSERR;
  }
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  return OK;
}

/* end of contentdatabase_mysql.c */
