/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file src/applications/afs/gtkui/download.c
 * @author Christian Grothoff
 * @author Igor Wronsky
 *
 * This is the code for handling download and other related
 * windows for the gtk+ client.
 **/

#include "config.h"

#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>

#include "util/logging.h"
#include "util/returnvalues.h"
#include "applications/afs/encoding/downloadutil.h"
#include "applications/afs/encoding/requestmanager.h"
#include "applications/afs/gtkui/helper.h"
#include "applications/afs/gtkui/download.h"
#include "applications/afs/gtkui/main.h"

/**
 * "delete_event" handler for the progressBarWindow. Sets the 
 * download state as aborted, and releases the wait semaphore for 
 * the thread awaiting download completion. Returns TRUE so that 
 * gtk doesn't destroy the window.
 *
 * @param widget here progressBarWindow
 * @param event dummy
 * @param dlm the accessor to the GTK window
 **/
static gint abortDownload(GtkWidget * widget,
                         GdkEvent * event,
                         DownloadModel * dlm) {
  // We don't know when Download() finds it suitable to
  // return, so we can't destroy anything here.
/*  LOG(LOG_WARNING, "DEBUG: abortDownload\n"); */
  dlm->downloadStatus = DOWNLOAD_ABORTED;
  LOG(LOG_DEBUG, 
      "DEBUG: Semaphore %d UP (abort)\n", 
      dlm->doneSem);
  gtk_widget_hide(dlm->progressBarWindow);
  SEMAPHORE_UP(dlm->doneSem);
  return(TRUE);
}

/**
 * This method is called by the download code to notify the user
 * interface of the download progress.
 *
 * @param stats the new statistical values
 * @param dlm the accessor to the GTK window
 **/
static void modelCallback(ProgressStats * stats,
			  DownloadModel * dlm) {
  double currentRetryAvg, averageBps;
  char scratch[256];

  if (_I_AM_DEAD == YES) {
    dlm->downloadStatus = DOWNLOAD_FAILED;
    if(dlm->doneSem != NULL) 
      SEMAPHORE_UP(dlm->doneSem);
    return;
  }

  if (dlm->downloadStatus != DOWNLOAD_OK) {
/*    LOG(LOG_WARNING, "DEBUG: Ret from modelC for %d\n", dlm); */
    return;
  }
  LOG(LOG_DEBUG, 
      "DEBUG: progress %d size %d pend %d sent %d"
      " currentr %d totalr %d ttl %d estimate %d\n",
      stats->progress, 
      stats->filesize,
      stats->requestsPending, 
      stats->requestsSent, 
      stats->currentRetries,
      stats->totalRetries,
      stats->currentTTL, 
      stats->duplicationEstimate);

  if(stats->requestsSent > 0)
    currentRetryAvg = 
      (double)stats->currentRetries / 
      (double)stats->requestsSent;
  else
    currentRetryAvg = 0;
    
  if(cronTime(NULL) - dlm->downloadStartTime > 0)
    averageBps =
      (double)(stats->progress) / 
      (((double)(cronTime(NULL) - dlm->downloadStartTime))/(double)cronSECONDS);
  else
    averageBps = 0;
 
  if(stats->progress == 0) {
    sprintf(scratch, 
 	    "ActiveReq: %4d AvgRetries/Active: %3.1f\nTotalRetries: %d\nFetching metablocks", 
	    stats->requestsSent,
  	    currentRetryAvg,
	    stats->totalRetries);
  } else {
    sprintf(scratch, 
 	    "ActiveReq: %4d AvgRetries/Active: %3.1f\nTotalRetries: %d\nAvgBPS: %8.3f", 
	    stats->requestsSent,
  	    currentRetryAvg,
	    stats->totalRetries,
	    averageBps);
  }
  /* update view */
  gdk_threads_enter();
  gtk_progress_set_value(GTK_PROGRESS(dlm->progressBar),
			 stats->progress);
  gtk_label_set_text(GTK_LABEL(dlm->progressRetries), 
		     scratch);
  gdk_threads_leave();

  /* the download is complete iff progress == filesize;
     if progress == filesize == 0, the download was
     aborted with an error */
  if (stats->progress == stats->filesize) {
    if (stats->filesize == 0) {
      LOG(LOG_DEBUG, 
	  "DEBUG: Semaphore %d UP (dl fail) to %d\n", 
          dlm->doneSem,
          dlm->doneSem->v);
      dlm->downloadStatus = DOWNLOAD_FAILED;
      SEMAPHORE_UP(dlm->doneSem);
    } else {
      LOG(LOG_DEBUG,
	  "DEBUG: Semaphore %d UP (dl done) to %d\n", 
          dlm->doneSem,
      	  dlm->doneSem->v);
      SEMAPHORE_UP(dlm->doneSem);
    }
  }
}

/**
 * Main function of the download thread. This function terminates
 * automagically once the download is complete or if the download
 * window is closed.  It is responsible for stopping the
 * requestmanager. The DownloadModel is freed by cascaded destruction
 * of the original saveAsWindow (which has been kept hidden).
 * 
 * @param dlm the download model
 **/
static void downloadFile_(DownloadModel * dlm) {
  GtkWidget * button;
  
  LOG(LOG_DEBUG,
      "DEBUG: Entering downloadFile_ for %s (%d), sem %d : %d\n", 
      dlm->fileName,
      dlm,
      dlm->doneSem,
      dlm->doneSem->v);

  /* initiate download */
  cronTime(&dlm->downloadStartTime);
  
  dlm->rm = downloadFile(&dlm->root.header.fileIdentifier.chk,
		         dlm->root.header.fileIdentifier.crc,
			 dlm->root.header.fileIdentifier.file_length,
			 dlm->fileName,
			 (ProgressModel)&modelCallback,
			 dlm);
  /* Wait here until download is complete or
     until the window is closed */
  LOG(LOG_DEBUG,
      "DEBUG: Passed, waiting for DL completion (%d), sem %d : %d\n",
      dlm,
      dlm->doneSem,
      dlm->doneSem->v);
  SEMAPHORE_DOWN(dlm->doneSem);
  LOG(LOG_DEBUG,
      "DEBUG: Passed, enter destroyRequestManager (%d), sem %d : %d\n",
      dlm,
      dlm->doneSem,
      dlm->doneSem->v);
  SEMAPHORE_FREE(dlm->doneSem);
  dlm->doneSem=NULL;

  /* Stop the RequestManager */
  destroyRequestManager(dlm->rm);
  
  LOG(LOG_DEBUG, 
      "DEBUG: Passed, gdk_threads_enter (%d)\n", dlm);
  
  gdk_threads_enter();
  
  LOG(LOG_DEBUG, 
      "DEBUG: Passed, reconn (%d)\n", dlm);
  
  /* Reconnect progressBarWindow to close the saveAsWindow */
  gtk_signal_disconnect_by_func(GTK_OBJECT(dlm->progressBarWindow), 
		     GTK_SIGNAL_FUNC(abortDownload), 
		     dlm);
  gtk_signal_connect(GTK_OBJECT(dlm->progressBarWindow), 
                     "delete_event",
	 	     GTK_SIGNAL_FUNC(deleteEvent),
	       	     NULL);
  gtk_signal_connect(GTK_OBJECT(dlm->progressBarWindow), 
                     "destroy",
	 	     GTK_SIGNAL_FUNC(destroyWidget),
	       	     dlm->saveAsWindow);
  
  /* Has the download window been closed (==abort) ? */
  if(dlm->downloadStatus == DOWNLOAD_ABORTED) {
    LOG(LOG_DEBUG, 
	"DEBUG: Download_aborted branch (%d)\n", dlm);
    gtk_widget_destroy(dlm->progressBarWindow);
    gdk_threads_leave();
    return;
  }

  LOG(LOG_DEBUG, 
      "DEBUG: Main branch (%d)\n", dlm);
  
  /* Show a button */
  if(dlm->downloadStatus == DOWNLOAD_FAILED) {
    gtk_progress_set_format_string(GTK_PROGRESS(dlm->progressBar),
		    	 	  "Download failed!");
    button = gtk_button_new_with_label("Bah.");
  } else {
    gtk_progress_set_format_string(GTK_PROGRESS(dlm->progressBar),
		  		 "Download complete");
    button = gtk_button_new_with_label("Yay!");
  }
  gtk_signal_connect(GTK_OBJECT(button), 
		     "clicked",
		     GTK_SIGNAL_FUNC(destroyWidget),
		     dlm->progressBarWindow);
  gtk_box_pack_start(GTK_BOX(dlm->progressBarWindowBox), 
		     button, 
		     FALSE, 
		     FALSE, 
		     0);
  gtk_widget_show(button);  
  LOG(LOG_DEBUG, "DEBUG: gdk_threads leaving (%d)\n", dlm);
  gdk_threads_leave();
  LOG(LOG_DEBUG,
      "DEBUG: DL left, downloadStatus=%d (%d)\n",
      dlm->downloadStatus,
      dlm);
}

/**
 * Open the download window and start the download of a file in the
 * background. The method executes during a signal handler, so a GTK
 * lock is not required to to GUI operations.
 *
 * @param dlm some information about what to download (startDownload
 *  will pass it to the download-thread)
 **/
void startDownload(DownloadModel * dlm) {
  GtkWidget * window;
  GtkWidget * box;
  GtkWidget * label;
  GtkObject * adjustment;
  char * fileNameRoot;
  int i;

  fileNameRoot = dlm->fileName;
  for (i=strlen(dlm->fileName)-1;i>=0;i--) {
    if (dlm->fileName[i] == '/') {
      fileNameRoot=&dlm->fileName[i+1];
      break;
    }
  }
  
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), fileNameRoot);
  box = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window), box);
  gtk_signal_connect(GTK_OBJECT(window), 
		     "delete_event",
		     GTK_SIGNAL_FUNC(abortDownload), 
		     dlm);
  gtk_container_set_border_width(GTK_CONTAINER (window), 10);
  /* progressbar */
  dlm->progressBar = gtk_progress_bar_new();
  dlm->progressBarWindow = window;
  dlm->progressBarWindowBox = box;
  gtk_progress_set_show_text(GTK_PROGRESS(dlm->progressBar),
			     1);
  gtk_progress_set_format_string(GTK_PROGRESS(dlm->progressBar),
				 (gchar*)"%v bytes downloaded");
  adjustment=gtk_adjustment_new(0,0,dlm->root.header.fileIdentifier.file_length,1,0,0);
  gtk_progress_set_adjustment(GTK_PROGRESS(dlm->progressBar),GTK_ADJUSTMENT(adjustment));
  gtk_box_pack_start(GTK_BOX(box), 
		     dlm->progressBar, TRUE, TRUE, 0);
  gtk_progress_bar_set_orientation(GTK_PROGRESS_BAR(dlm->progressBar),
				   GTK_PROGRESS_LEFT_TO_RIGHT);
  gtk_widget_show(dlm->progressBar);
  /* other progress stuff */
  label = gtk_label_new(" ");
  gtk_box_pack_start(GTK_BOX(dlm->progressBarWindowBox),
                        label, FALSE, FALSE, 0);
  gtk_widget_show(label);
  dlm->progressRetries=label; 
  
  dlm->doneSem = SEMAPHORE_NEW(0);
  /* main window */
  gtk_widget_show(box);
  gtk_widget_show(window);

  PTHREAD_CREATE(&dlm->downloadThread,
		 (void * (*)(void *)) downloadFile_,
		 dlm,
		 16 * 1024); 
  pthread_detach(dlm->downloadThread);
  
}

/* end of download.c */
