/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * ContentDatabse - block storage. MYSQL implementation.
 *
 * @author Igor Wronsky
 * @file include/applications/afs/database/contentdatabase_mysql.h
 **/

#ifndef CONTENTDATABASE_MYSQL_H
#define CONTENTDATABASE_MYSQL_H

#include "config.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <openssl/ripemd.h>
#ifndef SOMEBSD
 #ifndef CYGWIN
 #include <wordexp.h>
 #endif
#endif
#include <mysql/mysql.h>

#include "util/hashing.h"
#include "util/semaphore.h"
#include "applications/afs/database/contentdatabase.h"

/**
 * A mysql wrapper
 **/
typedef struct {
  MYSQL * dbf;
  Mutex DATABASE_Lock_;
} mysqlHandle;

/**
 * Initialize the storage module.
 * @param dir the name of the directory/file
 *        containing the content database
 **/
void * libInitContentDatabase(char * dir);

/**
 * Free the storage module.
 **/
void libDoneContentDatabase(mysqlHandle * dbh);

/**
 * Free space in the database by removing an entry.
 *
 * @param fn the key of the entry to remove
 * @return SYSERR on error, OK if ok.
 **/
int libUnlinkFromDB(mysqlHandle * dbh, 
		 HashCode160 * fn);

/**
 * Get the number of entries in the database.
 * @return SYSERR on error, otherwise the number of entries
 **/
int libCountContentEntries(mysqlHandle * dbh);

/**
 * Read the contents of a bucket to a buffer.
 *
 * @param fn the hashcode representing the entry
 * @param result the buffer to write the result to 
 *        (*result should be NULL, sufficient space is allocated)
 * @return the number of bytes read on success, SYSERR on failure
 **/ 
int libReadContent(mysqlHandle * dbh, 
	        HashCode160 *fn,
	        void ** result);

/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 *
 * @param fn the key for the entry
 * @param len the size of the block
 * @param block the data to store
 * @return SYSERR on error, OK if ok.
 **/
int libWriteContent(mysqlHandle * dbh,
		 HashCode160 * fn, 
		 int len,
		 void * block);

/**
 * Call a method for each key in the database and
 * call the callback method on it. 
 *
 * @param callback the callback method
 * @param data second argument to all callback calls
 * @return the number of items stored in the content database
 **/
int libForEachEntryInDatabase(mysqlHandle * dbh,
			   EntryCallback callback,
			   void * data);
 
#endif

/* end of contentdatabase_mysql.h */
