/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * This file contains the connection table which lists all the
 * current connections of the node with other hosts and buffers
 * outgoing packets to these hosts. The connection table also
 * contains state information such as sessionkeys, credibility and
 * the last time we had host activity.
 *
 * This code is responsible for exchanging a sessionkey with another
 * peer, grouping several messages into a larger packet, padding with
 * noise, encryption and deferred sending of these messages.  <p>
 *
 * The file is organized as follows:
 * 
 * a) includes
 * b) defines
 * c) typedefs
 * d) global variables
 * e) code
 * <p>
 *
 * @author Tzvetan Horozov
 * @author Christian Grothoff
 **/ 

/* ******************* includes ******************** */
#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "util/checksum.h"
#include "util/configuration.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/statistics.h"
#include "util/storage.h"
#include "util/symcipher.h"
#include "util/tcp_return.h"
#include "util/xmalloc.h"
#include "server/knownhosts.h"
#include "server/traffic.h"
#include "server/pingpong.h"
#include "server/handler.h"
#include "server/tcpserver.h"

/* **************** defines ************ */

/* tuning parameters */

#define DEBUG_CONNECTION NO

/**
 * If an attempt to establish a connection is not answered
 * within 60s, drop.
 **/
#define SECONDS_NOPINGPONG_DROP 60

/**
 * If an established connection is inactive for 5 minutes, 
 * drop.
 **/
#define SECONDS_INACTIVE_DROP 300

/**
 * After 4 minutes on an inactive connection, probe the other
 * node with a ping if we have achieved less than 50% of our
 * connectivity goal.
 **/
#define SECONDS_PINGATTEMPT 240

/* status constants */

#define STAT_DOWN             0
#define STAT_WAITING_FOR_PING 1
#define STAT_WAITING_FOR_PONG 2
#define STAT_UP               3

/* ******************** typedefs ******************* */

/**
 * Type of the linked list of send callbacks (to
 * implement a round-robbin invocation chain).
 **/
typedef struct SendCallbackList__ {
  /**
   * Minimum number of bytes that must be available
   * to call this callback.
   **/
  int minimumPadding;

  /**
   * The callback method.
   **/
  BufferFillCallback callback;

  /**
   * Did we say that this is a linked list?
   **/
  struct SendCallbackList__ * next;

} SendCallbackList;

/**
 * Record for state maintanance between scanHelperCount,
 * scanHelperSelect and scanForHosts.
 **/
typedef struct {
  int index;
  int matchCount;
  int costSelector;  
  HostIdentity match;
} IndexMatch;

typedef struct fENHWrap {
  PerNodeCallback method;
  void * arg;
} fENHWrap;

typedef struct {
  SEQUENCE_Message sequence;
  char variable[0];
} SequencedBuffer;

/**
 * Type of the connection table.
 **/
typedef struct BufferEntry_ {
  /** Session for the connection */
  Session session;
  /** How much do we trust the host? signed
      because that makes many operations that go 
      negative easier. Of course, negative trust makes
      no sense. */
  int trust;
  /** the current session key */
  SESSIONKEY skey;
  /** at which time was the sessionkey created (by whichever party) */
  time_t created;
  /** is this host alive? timestamp of the time of the last-active point */
  cron_t isAlive; 
  /** what is the timeout of a cron-job that will flush this buffer? */
  cron_t flushTime;
  /**  Status of the connection (STAT_XXX) **/
  unsigned int status;
  /** last sequence number received (!) on this connection (highest) */
  unsigned int lastSequenceNumberReceived;
  /** bit map indicating which of the 32 sequence numbers before the last
      were received (good for accepting out-of-order packets and
      estimating reliability of the connection) */
  unsigned int lastPacketsBitmap;
  /** priority for flushing/sending this buffer, max(priorities of packets in buffer) */
  unsigned int prio; /* what is the max. priority in this buffer */
 /** number of bytes used in the buffer */
  unsigned short len;
  /** contents of the buffer */
  SequencedBuffer * buffer;
  /** a hash collision overflow chain **/
  struct BufferEntry_ * overflowChain;

} BufferEntry;

/**
 * Type of a callback method on every buffer.
 * @param be the buffer entry
 * @param data context for callee
 **/
typedef void (*BufferEntryCallback)(BufferEntry * be,
				    void * data);

/* ***************** globals ********************** */

/**
 * The buffer containing all current connections.
 **/
static BufferEntry ** CONNECTION_buffer_;

/**
 * Size of the CONNECTION_buffer_
 **/
static unsigned int CONNECTION_MAX_HOSTS_;

/**
 * The DirContents array for scanning the hosts/ directory.
 **/
static int CONNECTION_currentActiveHosts;

/**
 * Experimental configuration: disable random padding of encrypted
 * messages.
 **/
static int disable_random_padding = NO;

/**
 * Send callbacks for making better use of noise padding...
 **/
static SendCallbackList * scl_nextHead;
static SendCallbackList * scl_nextTail;

/**
 * statistics handles
 **/
static int stat_number_of_connections;
static int stat_number_of_bytes_noise_send;
static int stat_number_of_bytes_send;
static int stat_number_of_bytes_received;
static int stat_sessionkeys_received;
static int stat_sessionkeys_verified;
static int stat_sessionkeys_transmitted;
static int stat_connections_shutdown;

/**
 * Lock for the connection module.
 **/
static Mutex lock;

/**
 * Where do we store trust information?
 **/
static char * trustDirectory;

/* ******************** CODE ********************* */

/** 
 * This initializes a given, preallocated BufferEntry structure.
 **/
static BufferEntry * initBufferEntry() {
  BufferEntry * be;

  be = (BufferEntry*) MALLOC(sizeof(BufferEntry));
  be->trust 
    = 0;
  be->isAlive 
    = 0;
  be->status 
    = STAT_DOWN;
  be->prio 
    = 0;
  be->buffer
    = NULL;
  be->len 
    = sizeof(SEQUENCE_Message);
  be->overflowChain
    = NULL;
  be->session.tsession
    = NULL;
  return be;
}

/**
 * From time to time, do a recount on how many hosts are connected.
 **/
static void cronCountConnections() {
  int act;
  int i;
  BufferEntry * root;
  BufferEntry * tmp;

  act = 0;
  MUTEX_LOCK(&lock);
  for (i=0;i< CONNECTION_MAX_HOSTS_;i++) {
    root = CONNECTION_buffer_[i];

    tmp = root;
    while(NULL != tmp) {
      if (tmp->status == STAT_UP)
        act++;
      tmp = tmp->overflowChain;
    }
    
  }
  CONNECTION_currentActiveHosts = act;
  statSet(stat_number_of_connections,
	  act);
  MUTEX_UNLOCK(&lock);
}


/**
 * Write host-infromation to a file - flush the buffer entry!
 * Assumes synchronized access.
 **/
static void flushHostCredit(BufferEntry * be,
			    void * unused) {
  HexName fil;
  char * fn;

  if ((be->trust & 0x80000000) == 0)
    return; /* unchanged */
  be->trust = be->trust & 0x7FFFFFFF;
  hash2hex(&be->session.sender.hashPubKey,
	   &fil);
  fn = MALLOC(strlen((char*)trustDirectory)+sizeof(HexName)+1);
  buildFileName(trustDirectory,
		&fil,
		fn);
  writeFile(fn, 
	    &be->trust, 
	    sizeof(unsigned int), 
	    "644");
  FREE(fn);
}

/**
 * Send a buffer; assumes that access is already synchronized.
 **/
static void sendBuffer(BufferEntry * be) {
  int crc;
  char * content;
  p2p_HEADER * part;
  int i;
  SendCallbackList * pos;

  if (be == NULL) {
    LOG(LOG_WARNING, 
	"WARNING: sendBuffer called with invalid arguments (NULL).\n");
    return;
  }  
  if (be->status == STAT_DOWN)
    return; /* status is down, nothing to send! (should at least be wait-for-ping/pong 
	       or up) */
  if (be->buffer == NULL) {
    LOG(LOG_FAILURE,
	"FAILURE: assertion failed: buffer NULL but status not down (%s:%d)\n",
	__FILE__, __LINE__);
    return; /* fail gracefully... */
  }
  be->flushTime = 0; /* flushing now! */
  if (be->len <= sizeof(SEQUENCE_Message))       
    return; /* nothing to send */  
  if (SYSERR == outgoingCheck(be->prio)) {
    be->prio 
      = 0;
    be->len 
      = sizeof(SEQUENCE_Message); /* too busy, truncate and drop! */
#if DEBUG_CONNECTION
    LOG(LOG_DEBUG,
	"DEBUG: policy (outgoing check) dropped message\n");
#endif
    return;
  }
  /* try callbacks */
  pos = scl_nextHead;
  while (pos != NULL) {
    if (pos->minimumPadding + be->len <= be->session.mtu) 
      be->len += pos->callback(&be->session.sender,
			       &((char*)be->buffer)[be->len],
			       be->session.mtu - be->len);
    pos = pos->next;
  }

  /* finally padd with noise */  

  if ( (be->len + sizeof(p2p_HEADER) <= be->session.mtu) &&
       (disable_random_padding == NO) ) {
    unsigned short noiseLen = be->session.mtu - be->len;

#if DEBUG_CONNECTION
    LOG(LOG_DEBUG,
	"DEBUG: padding with %d bytes of noise at %d\n",
	noiseLen, be->len);
#endif
    statChange(stat_number_of_bytes_noise_send,
	       noiseLen);
    part = (p2p_HEADER *) &((char*)be->buffer)[be->len];
    part->size 
      = htons(noiseLen);
    part->requestType 
      = htons(p2p_PROTO_NOISE);
    for (i=be->len+sizeof(p2p_HEADER)-sizeof(SEQUENCE_Message);
	 i < be->session.mtu-sizeof(SEQUENCE_Message);
	 i++)
      be->buffer->variable[i] = (char) rand();
    be->len 
      = be->session.mtu;
  }

  crc = crc32N(be->buffer, 
	       be->len);
  content = MALLOC(be->len);
  if (be->len == encryptBlock(be->buffer,
			      be->len,
			      &be->skey,
			      INITVALUE,
			      content)) {    
    statChange(stat_number_of_bytes_send,
	       be->len);
#if DEBUG_CONNECTION
    LOG(LOG_DEBUG,
	"DEBUG: calling transport layer to send %d bytes\n",
	be->len);
#endif
    transportSend(be->session.tsession,
		  content,
		  be->len,
		  YES, /* is encrypted */
		  crc);  
  } else
    LOG(LOG_FAILURE,
	"FAILURE: Encryption of message failed!\n");
  FREE(content);
  be->prio 
    = 0;
  be->len 
    = sizeof(SEQUENCE_Message); /* ok, mark buffer as send */
  be->buffer->sequence.sequenceNumber 
    = htonl(ntohl(be->buffer->sequence.sequenceNumber)+1);
}

static void sendBufferSynchronized(BufferEntry * be) {
  MUTEX_LOCK(&lock);
  sendBuffer(be);
  MUTEX_UNLOCK(&lock);
}

/**
 * Append a message to the current buffer. This method
 * assumes that the access to be is already synchronized.
 **/
static void appendToBuffer(BufferEntry * be,
			   p2p_HEADER * message,
			   unsigned int maxdelay) {
  unsigned short exSize;
  unsigned short newSize;
#if DEBUG_CONNECTION
  HexName hex;
  
  IFLOG(LOG_DEBUG,
	hash2hex(&be->session.sender.hashPubKey, 
		 &hex));
  LOG(LOG_DEBUG, 
      "DEBUG: adding message of size %d and type %d to buffer of host %s.\n",
      ntohs(message->size),
      ntohs(message->requestType),
      &hex);
#endif
  if (message == NULL) {
    LOG(LOG_WARNING, 
	"WARNING: appendToBuffer got garbage. Ignored.\n");
    return;
  }
  if (be->buffer == NULL) {
    LOG(LOG_FAILURE,
	"FAILURE: be->buffer NULL in append to buffer. Dropping message.\n");
    breakpoint();
    return;
  }

  newSize = ntohs(message->size);
  exSize = be->len;
  if (exSize + newSize > be->session.mtu) {
    if (newSize > be->session.mtu) {
      LOG(LOG_FAILURE, 
	  "FAILURE: appendToBuffer fatal error: exSize: %d newsize: %d BUFSIZE %d\n",
	  exSize, 
	  newSize, 
	  be->session.mtu);
      breakpoint();
      return;
    }
    sendBuffer(be); /* clean up*/
    exSize = sizeof(SEQUENCE_Message);
    if (exSize + newSize > be->session.mtu) {
      LOG(LOG_FAILURE, 
	  "FAILURE: appendToBuffer fatal error: exSize: %d newsize: %d BUFSIZE %d\n",
	  exSize, 
	  newSize, 
	  be->session.mtu);
      breakpoint();
      return;
    }
  }
  memcpy(&((char*)be->buffer)[exSize], 	 
	 message, newSize); 
  be->len = exSize + newSize;
  trafficSend(message,
	      &be->session.sender);


  /* flush the buffer after an average of maxdelay/2 seconds;
     if a message is already pending, do not schedule another
     cron job *except* if it's really urgent (instant send) */
  if (maxdelay == 0) {
    if (be->flushTime != 0)
      delCronJob((CronJob)&sendBufferSynchronized,
		 0,
		 be);
    sendBuffer(be);
  } else {
    cron_t now;
    int delta;

    delta = randomi(maxdelay+1);
    cronTime(&now);
#if DEBUG_CONNECTION
    LOG(LOG_DEBUG,
	"DEBUG: sending buffer deferred for at most %dms (%u)\n",
	delta,
	maxdelay);
#endif
    if (be->flushTime == 0) {
      be->flushTime = delta + now;
      addCronJob((CronJob)&sendBufferSynchronized,
		 delta,
		 0,
		 be);
    } else if (be->flushTime > delta + now) {
      delCronJob((CronJob)&sendBufferSynchronized,
		 0, 
		 be);
      be->flushTime = delta + now;
#if DEBUG_CONNECTION
      LOG(LOG_DEBUG,
	  "DEBUG: sending buffer deferred in at most %dms\n",
	  delta);
#endif
      addCronJob((CronJob)&sendBufferSynchronized,
		 delta,
		 0,
		 be);     
    }
  }
}

/**
 * Look for a host in the table. If the entry is there at the time of
 * checking, returns the entry.
 *
 * @returns the index of the host in the table
 **/
static BufferEntry * lookForHost(const HostIdentity * hostId) {
  BufferEntry * root;

  root = CONNECTION_buffer_[computeIndex(hostId)];
  while (root != NULL) {
    if (equalsHashCode160(&hostId->hashPubKey,
			  &root->session.sender.hashPubKey)) 
      return root;
    root = root->overflowChain;
  }
  return NULL;
}


/**
 * Read host-information from a file. The semaphore
 * of the given be must already be down.
 **/
static void initHostTrust(BufferEntry * be) {
  HexName fil;
  char * fn;

  hash2hex(&be->session.sender.hashPubKey,
	   &fil);
  fn = MALLOC(strlen((char*)trustDirectory)+sizeof(HexName)+1);
  buildFileName(trustDirectory, &fil, fn); 
  if (sizeof(unsigned int) !=
      readFile(fn, 
	       sizeof(unsigned int), 
	       &be->trust))
    be->trust = 0;
  FREE(fn);
}

/**
 * Force adding of a host to the buffer. If the node is already in the
 * table, the table entry is returned.  
 *
 * @param force if YES, drop another host from the table if the slot
 * is already in use. If NO, return NULL if the slot is busy.  
 * @return the table entry for the host (no keyexchange performed so far)
 **/
static BufferEntry * addHost(const HostIdentity * hostId,
			     int force) {
  BufferEntry * root;
  BufferEntry * prev;
  HexName hex;

  IFLOG(LOG_INFO,
	hash2hex(&hostId->hashPubKey, 
		 &hex));
#if DEBUG_CONNECTION
  LOG(LOG_INFO, 
      "INFO: Adding host %s to the connection table.\n",
      &hex);
#endif

  root = lookForHost(hostId);
  if (root != NULL)
    return root;

  root = CONNECTION_buffer_[computeIndex(hostId)];
  prev = NULL;
  while (NULL != root) {
    /* settle for entry in the linked list that is down */
    if ( (root->status == STAT_DOWN) ||
	 (equalsHashCode160(&hostId->hashPubKey,
			    &root->session.sender.hashPubKey)) ) 
      break;
    prev = root;
    root = root->overflowChain;
  }
  if (root == NULL) {
    root = initBufferEntry();
    if (prev == NULL)
      CONNECTION_buffer_[computeIndex(hostId)] = root;
    else
      prev->overflowChain = root;
  }
  memcpy(&root->session.sender,
	 hostId,
	 sizeof(HostIdentity));	 
  initHostTrust(root);
  return root;
}

/**
 * Perform an operation for all connected hosts.  The BufferEntry
 * structure is passed to the method.  No synchronization or other
 * checks are performed.
 *
 * @param method the method to invoke (NULL for couting only)
 * @param arg the second argument to the method
 * @return the number of connected hosts
 **/ 
static int forAllConnectedHosts(BufferEntryCallback method,
				void * arg) {
  unsigned int i;
  int count = 0;
  BufferEntry * be;
  
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    be = CONNECTION_buffer_[i];
    while (be != NULL) {
      if (be->status == STAT_UP) {
        if (method != NULL)
  	  method(be, arg);
        count++;
      }
      be = be->overflowChain;
    }
  }
  return count;
}


static void fENHCallback(BufferEntry * be,
			 void * arg) {
  fENHWrap * wrap;

  wrap = (fENHWrap*) arg;
  if (wrap->method != NULL)
    wrap->method(&be->session.sender,
		 wrap->arg);
}


/**
 * Here in this scanning for applicable hosts, we also want to take
 * the protocols into account and prefer "cheap" protocols,
 * i.e. protocols with a low overhead.
 **/
static void scanHelperCount(const HostIdentity * id,
			    const unsigned short proto,	
			    IndexMatch * im) {
  if (hostIdentityEquals(&myIdentity, id)) 
    return;
  if (computeIndex(id) != im->index)
    return;
  if (YES == isTransportAvailable(proto)) {
    im->matchCount++;  
    im->costSelector += transportGetCost(proto);
  }
#if DEBUG_CONNECTION
  else
    LOG(LOG_DEBUG,
	"DEBUG: transport %d is not available\n",
	proto);
#endif
}

static void scanHelperSelect(const HostIdentity * id,
			     const unsigned short proto,
			     IndexMatch * im) {
  if (hostIdentityEquals(&myIdentity, id)) 
    return;
  if (computeIndex(id) != im->index)
    return;
  if (YES == isTransportAvailable(proto)) {
    im->costSelector -= transportGetCost(proto);
    if ( (im->matchCount == 0) ||
	 (im->costSelector < 0) ) {
      memcpy(&im->match,
	     id,
	     sizeof(HostIdentity));
    }
    im->matchCount--;
  }
#if DEBUG_CONNECTION
  else
    LOG(LOG_DEBUG,
	"DEBUG: transport %d is not available\n",
	proto);
#endif
}

/**
 * Force creation of a new Session key for the given host. 
 * @param hostId the identity of the other host 
 * @param sk the SESSIONKEY to use
 * @param created the timestamp to use
 * @param ret the address where to write the signed session key, first unsigned short_SIZE byte give length
 * @return OK on success, SYSERR on failure
 **/
static int makeSessionKeySigned(HostIdentity * hostId,
				SESSIONKEY * sk,
				time_t created,
				SKEY_Message * ret) {
  HashCode160 keyHash;
  HostIdentity myId;
  HELO_Message * foreignHelo;
#if EXTRA_CHECKS
  HexName hostName;
#endif

  /* create and encrypt sessionkey */
  if (SYSERR == identity2Helo(hostId, 
			      ANY_PROTOCOL_NUMBER,
			      YES,
			      &foreignHelo)) {
    LOG(LOG_WARNING, 
	"WARNING: makeSesionKeySigned: cannot encrypt sessionkey, other host not known!\n");  
    return SYSERR; /* other host not known */  
  }
  if (SYSERR == encryptHostkey(sk,
			       sizeof(SESSIONKEY),
			       &foreignHelo->publicKey,
			       &ret->body.key)) {
    LOG(LOG_FAILURE, 
	"FAILURE: makeSesionKeySigned: encrypt failed!\n");
    FREE(foreignHelo);
    return SYSERR; /* encrypt failed */
  }
  FREE(foreignHelo);
  /* compute hash and sign hash */
  ret->body.creationTime = htonl(created);
  hash(&ret->body, 
       sizeof(RSAEncryptedData) + sizeof(time_t),
       &keyHash);
  if (SYSERR == signData(&keyHash, 
			 (unsigned short)sizeof(HashCode160),
			 &ret->body.signature)) {
    LOG(LOG_FAILURE, 
	"FAILURE: signData failed. Trying to continue.\n");
  }

  /* complete header */
  ret->header.size = htons(sizeof(SKEY_Message));
  ret->header.requestType = htons(p2p_PROTO_SKEY);
  /* verify signature/SKS */
  getHostIdentity(getPublicHostkey(),
		  &myId); 
#if EXTRA_CHECKS
  hash2hex(&hostId->hashPubKey, &hostName);
  if (SYSERR == verifySKS(&myId, ret)) {
    errexit("MakeSessionKey: generated sessionkey for host %s failed verification! exiting!\n",
	    &hostName);
  }
#endif
  return OK;
} 

/**
 * Perform a session key exchange for entry be.
 * First sends a HELO and then the new SKEY
 * (in two plaintext packets). When called, the 
 * semaphore of at the given index must already be down
 **/
static void exchangeKey(BufferEntry * be) {
  HexName hex;
  HELO_Message * helo;
  HELO_Message * targetHelo;
  SKEY_Message skey;
  unsigned short targetTransport;
  char * sendBuffer;

  IFLOG(LOG_DEBUG,
	hash2hex(&be->session.sender.hashPubKey, 
		 &hex));
#if DEBUG_CONNECTION
  LOG(LOG_DEBUG, 
      "DEBUG: Beginning of exchange key with %s\n",
	&hex);
#endif
  if (be->status != STAT_DOWN) 
    LOG(LOG_WARNING, 
	"WARNING: exchangeKey: status must be down for session key exchange!\n");

  makeSessionkey(&be->skey);
  time(&be->created); /* set creation time for session key! */
  if (SYSERR == makeSessionKeySigned(&be->session.sender,
				     &be->skey,
				     be->created,
				     &skey))
    return;
  be->isAlive = 0; /* wait a bit */
  be->status = STAT_WAITING_FOR_PING;
  be->lastSequenceNumberReceived = 0;
  be->lastPacketsBitmap = (unsigned int) -1;
  /* send tmp to target */
  if (SYSERR == identity2Helo(&be->session.sender,
			      ANY_PROTOCOL_NUMBER,
			      YES,
			      &targetHelo)) {
    return;
  }
  targetTransport = ntohs(targetHelo->protocol);
#if DEBUG_CONNECTION
  LOG(LOG_DEBUG,
      "DEBUG: identity2Helo returned HELO with protocol %d\n",
      targetTransport);
#endif
  if (SYSERR == transportCreateHELO(targetTransport,
				    &helo)) {
    be->status = STAT_DOWN;
    FREE(targetHelo);
    return;
  }
  if (SYSERR == transportConnect(targetHelo, /* callee frees except on SYSERR */
				 &be->session.tsession)) {
    be->status = STAT_DOWN;
    be->session.tsession = NULL;
    FREE(targetHelo);
    FREE(helo);
    return;
  } 
  targetHelo = NULL; /* ensure that we do not use the helo, 
			transportConnect is now the owner! */
  be->session.mtu = transportGetMTU(be->session.tsession->ttype);
  if (be->buffer != NULL)
    LOG(LOG_FAILURE,
	"FAILURE: expected be->buffer to be null in file %s:%d\n",
	__FILE__, __LINE__);
  be->buffer = MALLOC(be->session.mtu);
  be->buffer->sequence.header.size
    = htons(sizeof(SEQUENCE_Message));
  be->buffer->sequence.header.requestType
    = htons(p2p_PROTO_SEQUENCE);
  be->buffer->sequence.sequenceNumber
    = htonl(1);
  be->len = sizeof(SEQUENCE_Message);
  be->session.isEncrypted = NO;

  sendBuffer = MALLOC(HELO_Message_size(helo)+
		      sizeof(SKEY_Message));
  memcpy(sendBuffer,
	 helo,
	 HELO_Message_size(helo));
  memcpy(&sendBuffer[HELO_Message_size(helo)],
	 &skey,
	 sizeof(SKEY_Message));
  
  transportSend(be->session.tsession,
		sendBuffer,
		HELO_Message_size(helo)+sizeof(SKEY_Message),
		NO, /* not encrypted */
		crc32N(sendBuffer, HELO_Message_size(helo)+sizeof(SKEY_Message)));
  FREE(sendBuffer);
  FREE(helo);
  be->session.isEncrypted = YES;
  statChange(stat_sessionkeys_transmitted, 1);
}

/**
 * Look in the list for known hosts; pick a random host
 * of minimal transport cost for the hosttable at index 
 * index. When called, the 
 * mutex of at the given index must not be hold.
 **/
static void scanForHosts(unsigned int index) {
  BufferEntry * be;
  IndexMatch indexMatch;
  HexName hn;
  cron_t now;

#if DEBUG_CONNECTION
  LOG(LOG_CRON, 
      "CRON: Scanning for hosts (%d)\n",
      index);
#endif
  cronTime(&now);
  indexMatch.index = index;
  indexMatch.matchCount = 0;
  indexMatch.costSelector = 0;
  forEachHost((HostIterator)&scanHelperCount, 
	      now,
	      &indexMatch);
  if (indexMatch.matchCount == 0) 
    return;  
  LOG(LOG_CRON, 
      "CRON: Scanning for hosts (%d) found %d matching node identities.\n",
      index,
      indexMatch.matchCount);    
  /* indexMatch.matchCount 
    = randomi(indexMatch.matchCount); randomly choose host */
  indexMatch.costSelector
    = randomi(indexMatch.costSelector);
  memcpy(&indexMatch.match,
	 &myIdentity,
	 sizeof(HostIdentity));  
  forEachHost((HostIterator)&scanHelperSelect,
	      now,
	      &indexMatch);
  if (hostIdentityEquals(&myIdentity, 
			 &indexMatch.match)) {
    LOG(LOG_ERROR, 
	"ERROR: scanHelperSelect did not find selected match!?\n");
    return;
  }
  if (computeIndex(&indexMatch.match) != index) {
    LOG(LOG_ERROR, 
	"ERROR: index of host to add is wrong!\n");
    return;
  }
  IFLOG(LOG_DEBUG,
	hash2hex(&indexMatch.match.hashPubKey,
		 &hn));
#if DEBUG_CONNECTION
  LOG(LOG_DEBUG, 
      "DEBUG: attempting to connect to %s using slot %d.\n",
      &hn, 
      index);
#endif
  be = addHost(&indexMatch.match, 
	       NO); /* NO or YES should not matter here since we know the slot is empty */
  if (be != NULL) {
    if (be->status == STAT_DOWN) {
      blacklistHost(&be->session.sender,
		    CONNECTION_currentActiveHosts); 
      /* we're trying now, don't try again too soon */
      exchangeKey(be);
    }
  }
}


/**
 * Check if the buffer is up (we got a PONG), if not, repeat the PING 
 **/
static void checkAndPing(BufferEntry * be) {
  HostIdentity * data;

  LOG(LOG_CRON, 
      "CRON: enter checkAndPing\n");
  MUTEX_LOCK(&lock);
  if ( (be->status != STAT_UP) &&
       (be->buffer != NULL) ) {
    PINGPONG_Message pmsg;

    data = MALLOC(sizeof(HostIdentity));
    memcpy(data,
	   &be->session.sender,
	   sizeof(HostIdentity));
    if (OK == pingAction(&be->session.sender,
			 (CronJob) &notifyPONG,
			 data,
			 &pmsg)) {
      /* the fewer connections, the more important is the pinging */
      be->prio += getConnectPriority();
      /* force flushing NOW */
      appendToBuffer(be, &pmsg.header, 0);
    } else {
      FREE(data);
      LOG(LOG_WARNING,
	  "WARNING: could not send checking ping, ping buffer full\n");
    }
  }
  MUTEX_UNLOCK(&lock);
  LOG(LOG_CRON, 
      "CRON: exit checkAndPing\n");
}

/**
 * Check if the buffer is up, if not, revert the state to 'DOWN'.
 **/
static void checkAndGiveup(BufferEntry * be) {
  LOG(LOG_CRON, 
      "CRON: enter checkAndGiveup\n");
  MUTEX_LOCK(&lock);
  if (be->status != STAT_UP) {
    be->status = STAT_DOWN;
    if (be->session.tsession != NULL)
      transportDisconnect(be->session.tsession);
    be->session.mtu = 0;
    be->session.tsession = NULL;
    FREENONNULL(be->buffer);
    be->buffer = NULL;
  }
  MUTEX_UNLOCK(&lock);
  LOG(LOG_CRON, 
      "CRON: exit checkAndGiveup\n");
}

/**
 * Shutdown the connection.
 * Send a HANGUP message to the other side and
 * mark the sessionkey as dead.
 **/
static void shutdownConnection(BufferEntry * be) {
  HANGUP_Message hangup;

  if (be->status == STAT_DOWN)
    return; /* nothing to do */
  if (be->status == STAT_UP) {
    /* send HANGUP */
    hangup.header.requestType = htons(p2p_PROTO_HANGUP);
    hangup.header.size = htons(sizeof(HANGUP_Message));
    getHostIdentity(getPublicHostkey(),
		    &hangup.sender);
    appendToBuffer(be, 
		   &hangup.header, 
		   0); /* 0: send NOW */
  }
  be->created = 0;
  be->status = STAT_DOWN;
  if (be->session.tsession != NULL) {
    transportDisconnect(be->session.tsession);
    be->session.tsession = NULL;
  }    
  statChange(stat_connections_shutdown, 1);
  FREENONNULL(be->buffer);
  be->buffer = NULL;
  delCronJob((CronJob) &sendBufferSynchronized,
	     0,
	     be);
  delCronJob((CronJob)&checkAndPing,
	     0,
	     be);
  delCronJob((CronJob)&checkAndGiveup,
	     0,
	     be);
}

/**
 * Call this method periodically to decrease liveness of hosts.
 **/
static void cronDecreaseLiveness(void * unused) {  
  static int lastLivenessHost = 0;
  BufferEntry * root;
  BufferEntry * prev;
  BufferEntry * tmp;
  cron_t now;
  
  /* Find the correct host */
  root = CONNECTION_buffer_[lastLivenessHost];
  cronTime(&now);    
  MUTEX_LOCK(&lock);
  prev = NULL;
  while (NULL != root) {
    /* is this the host? */
    switch (root->status) {
    case STAT_DOWN:
      if (prev == NULL)
	CONNECTION_buffer_[lastLivenessHost] = root->overflowChain;
      else
	prev->overflowChain = root->overflowChain;
      tmp = root;
      root = root->overflowChain;
      FREE(tmp);
      continue;
    case STAT_UP:
      if (now - root->isAlive > SECONDS_INACTIVE_DROP * cronSECONDS) {
	/* switch state form UP to DOWN: too much inactivity */
	shutdownConnection(root);
	whitelistHost(&root->session.sender); /* the host may still be worth trying again soon */
	CONNECTION_currentActiveHosts--;
	statChange(stat_number_of_connections,
		   -1);
      } else if ( (CONNECTION_currentActiveHosts < CONNECTION_MAX_HOSTS_/2) &&
		  (now - root->isAlive > SECONDS_PINGATTEMPT * cronSECONDS) ){
	/* if we have less than 50% of the number of connections
	   that we would like to have, try ping-ing the other side
	   to keep the connection open instead of hanging up */
	PINGPONG_Message pmsg;
	HostIdentity * hi = MALLOC(sizeof(HostIdentity));
	memcpy(hi, 
	       &root->session.sender, 
	       sizeof(HostIdentity));
	if (OK == pingAction(&root->session.sender,
			     (CronJob)&notifyPING,
			     hi,
			     &pmsg)) {
	  appendToBuffer(root, 
			 &pmsg.header, 
			 (SECONDS_INACTIVE_DROP - SECONDS_PINGATTEMPT) * cronSECONDS / 2);
	} else
	  FREE(hi);
      }
      break;
    case STAT_WAITING_FOR_PING:
    case STAT_WAITING_FOR_PONG:
      if (now - root->isAlive > SECONDS_NOPINGPONG_DROP * cronSECONDS) 
	shutdownConnection(root);      
      break;
    default:
      LOG(LOG_WARNING,
	  "WARNING: unknown type of connection status: %d\n",
	  root->status);
      break; /* do nothing */
    } /* end of switch */ 
    prev = root;
    root = root->overflowChain;
  } /* end of while */
  if (CONNECTION_buffer_[lastLivenessHost] == NULL)
    scanForHosts(lastLivenessHost);

  lastLivenessHost++;
  if(lastLivenessHost >= CONNECTION_MAX_HOSTS_)
    lastLivenessHost = 0;
  MUTEX_UNLOCK(&lock);
}

/**
 * Check the sequence number. Updates the sequence
 * number as a side-effect.
 * @returns OK if ok, SYSERR if not.
 **/
static int checkSequenceNumber(HostIdentity * sender,
			       p2p_HEADER * msg) {
  SEQUENCE_Message * smsg;
  BufferEntry * be;
  int res;
  unsigned int sequenceNumber;
  
  if (ntohs(msg->size) != sizeof(SEQUENCE_Message)) {
    LOG(LOG_WARNING,
	"WARNING: sequence message received has wrong size %d\n",
	ntohs(msg->size));
    return SYSERR;
  }
  smsg = (SEQUENCE_Message*) msg;
  sequenceNumber = ntohl(smsg->sequenceNumber);
  MUTEX_LOCK(&lock);
  be = lookForHost(sender);
  if (be == NULL) {
    LOG(LOG_FAILURE,
	"FAILURE: sequence message received for node that is not connected!?\n");
    MUTEX_UNLOCK(&lock);
    return SYSERR; /* host not found */
  }
  res = OK;
  if (be->lastSequenceNumberReceived >= sequenceNumber) {
    unsigned int rotbit = 1;
    if ( (be->lastSequenceNumberReceived - sequenceNumber <= 32) && 
	 (be->lastSequenceNumberReceived != sequenceNumber) ) {
      rotbit = rotbit << (be->lastSequenceNumberReceived - sequenceNumber - 1);
      if ( (be->lastPacketsBitmap & rotbit) == 0) {
	res = OK;
	be->lastPacketsBitmap |= rotbit;
      } else
	res = SYSERR;
    } else
      res = SYSERR;
    if (res == SYSERR) {
      LOG(LOG_WARNING,
	  "WARNING: Invalid sequence number %u < %u, dropping rest of packet\n",
	  sequenceNumber, 
	  be->lastSequenceNumberReceived);
    }    
  } else {    
    be->lastPacketsBitmap = 
      be->lastPacketsBitmap << (sequenceNumber - be->lastSequenceNumberReceived);
    be->lastSequenceNumberReceived = sequenceNumber;
  }
  MUTEX_UNLOCK(&lock);
  return res;
}

/**
 * Check the sequence number. Updates the sequence
 * number as a side-effect.
 * @returns OK if ok, SYSERR if not.
 **/
static int processGetConnectionCountRequest(GNUNET_TCP_SOCKET * client,
					    CS_HEADER * msg) {
  if (ntohs(msg->size) != sizeof(CS_HEADER)) {
    LOG(LOG_WARNING,
	"WARNING: request received from TCP has wrong size %d\n",
	ntohs(msg->size));
    return SYSERR;
  }
  return sendTCPResult(client,
		       CONNECTION_currentActiveHosts);
}


/**
 * Handler for processing HANGUP.
 **/
static int handleHANGUP(HostIdentity * sender,
			p2p_HEADER * msg) {
  BufferEntry * be;
  HexName hex;

  if (ntohs(msg->size) != sizeof(HANGUP_Message))
    return SYSERR;
  if (!hostIdentityEquals(sender,
			  &((HANGUP_Message*)msg)->sender))
    return SYSERR;
  IFLOG(LOG_INFO,
	hash2hex(&sender->hashPubKey, 
		 &hex));
#if DEBUG_CONNECTION
  LOG(LOG_INFO,
      "INFO: received HANGUP from %s\n",
      &hex);
#endif
  MUTEX_LOCK(&lock);
  be = lookForHost(sender);
  if (be == NULL) {
    MUTEX_UNLOCK(&lock);
    return SYSERR;
  }
  shutdownConnection(be);
  MUTEX_UNLOCK(&lock);
  return OK;
}

/**
 * @return SYSERR if invalid, OK if valid
 **/
static int verifySKS(const HostIdentity * hostId,
		     SKEY_Message * sks) {
  HashCode160 keyHash;
  HexName hostName;
  HELO_Message * helo;
  
  if ( (sks == NULL) ||
       (hostId == NULL)) {
    IFLOG(LOG_WARNING,
	  hash2hex(&hostId->hashPubKey, 
		   &hostName));
    LOG(LOG_WARNING, 
	"WARNING: VerifySKS: invalid sessionkey suggestion for host %s\n",
	&hostName);
    return SYSERR;
  }
  
  if (SYSERR == identity2Helo(hostId, 
			      ANY_PROTOCOL_NUMBER,
			      YES,
			      &helo)) {
    IFLOG(LOG_WARNING,
	  hash2hex(&hostId->hashPubKey, 
		   &hostName));
    LOG(LOG_WARNING, 
	"WARNING: verifySKS: host %s for sessionkey exchange not known\n",
	&hostName);
    return SYSERR;
  }
  /* verify signature */
  hash(&sks->body, 
       sizeof(RSAEncryptedData) + sizeof(time_t),
       &keyHash);
  if (!verifySig(&keyHash,
		 sizeof(HashCode160),
		 &sks->body.signature, 
		 &helo->publicKey)) {
    LOG(LOG_WARNING, 
	"WARNING: verifySKS: received SessionKey with bad signature!\n");
    FREE(helo);
    return SYSERR; /*reject!*/
  }
  FREE(helo);
  statChange(stat_sessionkeys_verified, 1);
  return OK; /* ok */
}

/**
 * Call once in a while :-)
 **/
static void cronFlushTrustBuffer(void * unused) {
#if DEBUG_CONNECTION
  LOG(LOG_CRON,
      "CRON: enter cronFlushTrustBuffer\n");
#endif
  MUTEX_LOCK(&lock);
  forAllConnectedHosts(&flushHostCredit, unused);
  MUTEX_UNLOCK(&lock);
#if DEBUG_CONNECTION
  LOG(LOG_CRON, 
      "CRON: exit cronFlushTrustBuffer\n");
#endif
}

/**
 * How important is it at the moment to establish more connections?
 **/
int getConnectPriority() {
  if (CONNECTION_MAX_HOSTS_ > CONNECTION_currentActiveHosts)
    return (CONNECTION_MAX_HOSTS_ - CONNECTION_currentActiveHosts)*5;
  else
    return 0;
}

/**
 * Consider switching the transport mechanism used for contacting
 * the given node. This function is called when the handler handles
 * an encrypted connection. For example, if we are sending SMTP
 * messages to a node behind a NAT box, but that node has established
 * a TCP connection to us, it might just be better to send replies
 * on that TCP connection instead of keeping SMTP going.<p>
 *
 * We can only successfully takeover if the transport is bidirectional
 * (can be associated). It also only makes sense if the cost is lower.
 * This method checks both.
 *
 * @param tsession the transport session that is for grabs
 * @param sender the identity of the other node
 **/
void considerTakeover(TSession * tsession,
		      HostIdentity * sender) {
  BufferEntry * be;

  if (tsession == NULL)
    return;
  MUTEX_LOCK(&lock);
  be = lookForHost(sender);
  if (be != NULL) {
    if (be->status != STAT_DOWN) {
      unsigned int cost = -1;
      if (be->session.tsession != NULL)
	cost = transportGetCost(be->session.tsession->ttype);
      /* Question: doesn't this always do takeover in tcp/udp
	 case, which have the same costs? Should it? -IW 

	 Answer: this will always switch to TCP in the long run (if
	 that is possible) since udpAssociate always
	 returns SYSERR. This is intended since for long-running
	 sessions, TCP is the better choice. UDP is only better for
	 sending very few messages (e.g. attempting an initial exchange
	 to get to know each other). See also transport paper and the
	 data on throughput. - CG
      */
      if (transportGetCost(tsession->ttype) >= cost) {
	if (transportAssociate(tsession) == OK) {
	  transportDisconnect(be->session.tsession);
	  be->session.tsession = tsession;
	}
      } /* end if cheaper AND possible */
    } /* end if connected */
  }
  MUTEX_UNLOCK(&lock);
}


/**
 * Accept a session-key that has been sent by another host.
 * The other host must be known (public key)
 * @param hostId the identity of the sender host
 * @param sessionkeySigned the session key that was "negotiated"
 **/
int acceptSessionKey(HostIdentity * sender,
		     TSession * tsession,
		     p2p_HEADER * msg) {
  BufferEntry * be;
  HexName hostName;
  SESSIONKEY key;
  SKEY_Message * sessionkeySigned;

  statChange(stat_sessionkeys_received, 1);
  if (ntohs(msg->size) != sizeof(SKEY_Message))
    return SYSERR;
  sessionkeySigned = (SKEY_Message *) msg;
  IFLOG(LOG_DEBUG,
	hash2hex(&sender->hashPubKey,
		 &hostName));
#if DEBUG_CONNECTION
  LOG(LOG_DEBUG, 
      "DEBUG: Received sessionkey from host %s.\n",
      &hostName);
#endif
  if (SYSERR == verifySKS(sender, 
			  sessionkeySigned)) {
    IFLOG(LOG_WARNING,
	  hash2hex(&sender->hashPubKey,
		   &hostName));
    LOG(LOG_WARNING, 
	"WARNING: session key from %s failed verification, ignored!\n", 
	&hostName);
    return SYSERR;  /* rejected */
  }

  /* prepare file */
  if (sizeof(SESSIONKEY) != 
      decryptData(&sessionkeySigned->body.key,
		  &key,
		  sizeof(SESSIONKEY))) {
    IFLOG(LOG_WARNING,
	  hash2hex(&sender->hashPubKey,
		   &hostName));
    LOG(LOG_WARNING, 
	"WARNING: SKEY rejected from host %s\n",
	&hostName);
    return SYSERR;
  }

  MUTEX_LOCK(&lock);
  be = lookForHost(sender);
  if (be == NULL) 
    be = addHost(sender, NO);
  if (be == NULL) {
    IFLOG(LOG_WARNING,
	  hash2hex(&sender->hashPubKey,
		   &hostName));
    LOG(LOG_INFO, 
	"INFO: Session key exchange denied, slot busy.\n",
	&hostName);
   MUTEX_UNLOCK(&lock);
   return SYSERR;
  }

  if (be->created > ntohl(sessionkeySigned->body.creationTime)) {
    LOG(LOG_INFO, 
	"INFO: key dropped, we've send or received a more recent key!\n");
    MUTEX_UNLOCK(&lock);
    return SYSERR;
  }

  /* if we have another connection established with the node,
     shut it down! */
  if (be->session.tsession != NULL) 
    shutdownConnection(be);

  /* try to associate with an existing connection (if the other side
     contacted us with a bi-directional protocol, if that fails, establish
     our own connection */
  if (SYSERR == transportAssociate(tsession)) {
    HELO_Message * helo;

    tsession = NULL;
    if (SYSERR == identity2Helo(sender,
				ANY_PROTOCOL_NUMBER,
				NO,
				&helo)) {
      LOG(LOG_WARNING, 
	  "WARNING: sessionkey received, but could not find transport to reply\n");
     MUTEX_UNLOCK(&lock);
     return SYSERR;
    }
    if (SYSERR == transportConnect(helo,
				   &tsession)) {
       LOG(LOG_WARNING, 
	  "WARNING: sessionkey received, but transport failed to connect\n");
       FREE(helo);
      MUTEX_UNLOCK(&lock);
      return SYSERR;
    }
  }
  /* ok, everything set, let's change the state to waiting for ping
     and initialize the buffer entry */
  memcpy(&be->skey,
	 &key,
	 SESSIONKEY_LEN);
  be->session.tsession 
    = tsession;
  be->session.isEncrypted 
    = YES;
  be->session.mtu 
    = transportGetMTU(be->session.tsession->ttype);
  be->created 
    = ntohl(sessionkeySigned->body.creationTime);
  be->status 
    = STAT_WAITING_FOR_PONG;
  be->lastSequenceNumberReceived 
    = 0;
  be->lastPacketsBitmap 
    = (unsigned int) -1; /* all bits set */
  if (be->buffer != NULL)
    LOG(LOG_FAILURE,
	"FAILURE: expected be->buffer to be null in file %s:%d\n",
	__FILE__, __LINE__);
  be->buffer 
    = MALLOC(be->session.mtu);
  be->buffer->sequence.header.size
    = htons(sizeof(SEQUENCE_Message));
  be->buffer->sequence.header.requestType
    = htons(p2p_PROTO_SEQUENCE);
  be->buffer->sequence.sequenceNumber
    = htonl(1);
  be->len 
    = sizeof(SEQUENCE_Message);

#if DEBUG_CONNECTION
  LOG(LOG_DEBUG, 
      "DEBUG: SKEY exchange - sending encrypted ping\n");
#endif
  MUTEX_UNLOCK(&lock);
  checkAndPing(be);
  delCronJob((CronJob)&checkAndPing,
	     0, 
	     be);
  delCronJob((CronJob)&checkAndGiveup,
	     0, 
	     be);
  addCronJob((CronJob)&checkAndPing,
	     90 * cronSECONDS,
	     0, 
	     be);
  addCronJob((CronJob)&checkAndPing,
	     3 * cronMINUTES, 
	     0, 
	     be);
  addCronJob((CronJob)&checkAndGiveup,
	     5 * cronMINUTES,
	     0, 
	     be);
  return OK;
}

/**
 * Initialize this module.
 **/
void initConnection() {
  unsigned int i;

  stat_sessionkeys_received 
    = statHandle("# sessionkeys received");
  stat_sessionkeys_verified 
    = statHandle("# valid sessionkeys received");
  stat_sessionkeys_transmitted
    = statHandle("# sessionkeys sent");
  stat_connections_shutdown
    = statHandle("# connections shutdown");
  stat_number_of_connections 
    = statHandle("# currently connected nodes");
  stat_number_of_bytes_noise_send
    = statHandle("# bytes noise sent");
  stat_number_of_bytes_send
    = statHandle("# encrypted bytes sent");
  stat_number_of_bytes_received
    = statHandle("# bytes decrypted");
  scl_nextHead 
    = NULL;
  scl_nextTail 
    = NULL;
  MUTEX_CREATE_RECURSIVE(&lock);
  CONNECTION_currentActiveHosts 
    = 0;
  CONNECTION_MAX_HOSTS_ 
    = getConfigurationInt("GNUNETD",
			  "MAXCONNECT");
  if (CONNECTION_MAX_HOSTS_ < 4)
    CONNECTION_MAX_HOSTS_ = 4;
   i = 1;
  while (i< CONNECTION_MAX_HOSTS_)
    i*=2;
  CONNECTION_MAX_HOSTS_ = i; /* make sure it's a power of 2 */

  CONNECTION_buffer_ 
    = (BufferEntry**) MALLOC(sizeof(BufferEntry*)*CONNECTION_MAX_HOSTS_);
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) 
    CONNECTION_buffer_[i] = NULL;

  registerp2pHandler(p2p_PROTO_SEQUENCE,
		     &checkSequenceNumber);
  registerp2pHandler(p2p_PROTO_HANGUP,
		     &handleHANGUP);
  registerCSHandler(CS_PROTO_CLIENT_COUNT,
		    &processGetConnectionCountRequest);
  addCronJob(&cronCountConnections,
	     1 * cronMINUTES,
	     30 * cronSECONDS,
	     NULL);
  addCronJob(&cronDecreaseLiveness,
	     1 * cronSECONDS, 
	     cronMINUTES / CONNECTION_MAX_HOSTS_,
	     NULL); /* Liveness in 1 minute intervals */  
  disable_random_padding = testConfigurationString("GNUNETD-EXPERIMENTAL",
						   "PADDING",
						   "NO");
  trustDirectory = getFileName("GNUNETD",
			       "TRUSTDIR",
			       "Configuration file must specify directory for trust in section %s under %s.\n");
  mkdirp(trustDirectory);
  addCronJob(&cronFlushTrustBuffer,
	     5 * cronMINUTES,
	     5 * cronMINUTES,
	     NULL); 
}


/**
 * Shutdown the connection module.
 **/
void doneConnection() {
  int i;
  BufferEntry * be;
  SendCallbackList * scl;
  
  delCronJob(&cronFlushTrustBuffer,
	     5 * cronMINUTES,
	     NULL);
  unregisterCSHandler(CS_PROTO_CLIENT_COUNT,
		      &processGetConnectionCountRequest);
  unregisterp2pHandler(p2p_PROTO_SEQUENCE,
		     &checkSequenceNumber);
  delCronJob(&cronCountConnections,
	     30 * cronSECONDS, 
	     NULL);
  delCronJob(&cronDecreaseLiveness,
	     cronMINUTES / CONNECTION_MAX_HOSTS_, 
	     NULL); 
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    BufferEntry * prev;

    prev = NULL;
    be = CONNECTION_buffer_[i];
    while (be != NULL) {
      shutdownConnection(be);
      flushHostCredit(be, NULL);    
      delCronJob((CronJob) &sendBufferSynchronized,
		 0,
		 be);
      delCronJob((CronJob)&checkAndPing,
		 0, 
	       be);
      delCronJob((CronJob)&checkAndGiveup,
		 0,
		 be);    
      prev = be;
      be = be->overflowChain;
      FREE(prev);
    }
  } 
  MUTEX_DESTROY(&lock);
  FREE(CONNECTION_buffer_);
  FREE(trustDirectory);
  trustDirectory = NULL;
  CONNECTION_buffer_ = NULL;
  while (scl_nextHead != NULL) {
    scl = scl_nextHead;
    scl_nextHead = scl_nextHead->next;
    FREE(scl);
  }
  scl_nextTail = NULL;

}

/**
 * Increase the host credit by a value. 
 *
 * @param hostId is the identity of the host
 * @param value is the int value by which the host credit is to be increased or
 *        decreased
 * @returns the actual change in trust (positive or negative)
 */
unsigned int changeHostCredit(const HostIdentity * hostId, 
			      int value){
  BufferEntry * be;

  MUTEX_LOCK(&lock);
  be = lookForHost(hostId);
  if (be == NULL) 
    return 0; /* not connected! */
  if ( (be->trust & 0x7FFFFFF) + value < 0) {
    value = - be->trust;
    be->trust = 0x80000000;
  } else {
    be->trust += value;
    be->trust = be->trust | 0x80000000;
  }
  MUTEX_UNLOCK(&lock);
  return value;
}

/**
 * Obtain the trust record of the host.
 **/
unsigned int getHostCredit(const HostIdentity * hostId) {
  BufferEntry * be;
  unsigned int trust;

  MUTEX_LOCK(&lock);
  be = lookForHost(hostId);
  if (be == NULL) {
    MUTEX_UNLOCK(&lock);
    return 0;
  }
  trust = be->trust;
  MUTEX_UNLOCK(&lock);
  return trust & 0x7FFFFFFF;
}

/**
 * Wrapper around forAllConnectedHosts...
 **/
int forEachConnectedNode(PerNodeCallback method,
			 void * arg) {
  fENHWrap wrap;
  int ret;

  wrap.method = method;
  wrap.arg = arg;
  MUTEX_LOCK(&lock);
  ret = forAllConnectedHosts(&fENHCallback,
			     &wrap);
  MUTEX_UNLOCK(&lock);
  return ret;
}

/**
 * Print the contents of the connection buffer.
 **/
void printConnectionBuffer() {
  int i;
  BufferEntry * tmp;
  HexName hostName;
  HexName skey;
  unsigned short ttype;

  MUTEX_LOCK(&lock);
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    tmp = CONNECTION_buffer_[i];
    while (tmp != NULL) {
      if (tmp->status != STAT_DOWN) {
        IFLOG(LOG_MESSAGE,
  	      hash2hex(&tmp->session.sender.hashPubKey, 
  	  	       &hostName));
	IFLOG(LOG_MESSAGE,
	      hash2hex((HashCode160*) &tmp->skey,
		       &skey));
	ttype = 0;
	if (tmp->session.tsession != NULL)
	  ttype = tmp->session.tsession->ttype;
	LOG(LOG_MESSAGE,
  	    "CONNECTION-TABLE: %3d-%1d-%2d-%6u-%4ds (of %ds)-%4u: %20s-%16s\n",
	    i,
	    tmp->status,
	    ttype,
	    tmp->trust & 0x7FFFFFFF, 
	    (int) ((cronTime(NULL) - tmp->isAlive)/cronSECONDS), 
	    SECONDS_INACTIVE_DROP,
	    tmp->len,
	    &hostName,
	    &skey);
      }
      tmp = tmp->overflowChain;
    }
  }
  MUTEX_UNLOCK(&lock);
}

/**
 * Register a callback method that should be invoked whenever a
 * message is about to be send that has more than minimumPadding bytes
 * left before maxing out the MTU. The callback method can then be
 * used to add additional content to the message (instead of the
 * random noise that is added by otherwise). Note that if the MTU is 0
 * (for streams), the callback method will always be called with
 * padding set to the maximum number of bytes left in the buffer
 * allocated for the send.
 *
 * @param minimumPadding how large must the padding be in order
 *   to call this method?
 * @param callback the method to invoke. The receiver is the
 *   receiver of the message, position is the reference to the
 *   first unused position in the buffer where GNUnet is building
 *   the message, padding is the number of bytes left in that buffer.
 *   The callback method must return the number of bytes written to
 *   that buffer (must be a positive number).
 * @return OK if the handler was registered, SYSERR on error
 **/
int registerSendCallback(const int minimumPadding,
			 BufferFillCallback callback) {
  SendCallbackList * scl;

  scl = MALLOC(sizeof(SendCallbackList));
  scl->minimumPadding = minimumPadding;
  scl->callback = callback;
  scl->next = NULL;
  MUTEX_LOCK(&lock);
  if (scl_nextTail == NULL) {
    scl_nextHead = scl;
    scl_nextTail = scl;
  } else {
    scl_nextTail->next = scl;
    scl_nextTail = scl;
  }
  MUTEX_UNLOCK(&lock);
  return OK;  
}
 
/**
 * Unregister a handler that was registered with registerSendCallback.
 * @return OK if the handler was removed, SYSERR on error
 **/
int unregisterSendCallback(const int minimumPadding,
			   BufferFillCallback callback) {
  SendCallbackList * pos;
  SendCallbackList * prev;

  prev = NULL;
  MUTEX_LOCK(&lock);
  pos = scl_nextHead;
  while (pos != NULL) {
    if ( (pos->callback == callback) &&
	 (pos->minimumPadding == minimumPadding) ) {
      if (prev == NULL) 
	scl_nextHead = pos->next;
      else
	prev->next = pos->next;
      if (scl_nextTail == pos)
	scl_nextTail = prev;
      FREE(pos);
      MUTEX_UNLOCK(&lock);
      return OK;     
    }
    prev = pos;
    pos = pos->next;
  }
  MUTEX_UNLOCK(&lock);
  return SYSERR;
}

/**
 * We received a sign of life from this host. 
 **/
void notifyPONG(const HostIdentity * hostId) {
  BufferEntry * be;
  HexName hex;

  MUTEX_LOCK(&lock);
  be = lookForHost(hostId);
  if (be != NULL) {
    switch (be->status) {
    case STAT_DOWN:
      break;
    case STAT_WAITING_FOR_PING:
      /* wrong message, ignore */
      break;
    case STAT_WAITING_FOR_PONG:
      be->status = STAT_UP;
      CONNECTION_currentActiveHosts++;
      cronTime(&be->isAlive);
      statChange(stat_number_of_connections,
		 1);
      IFLOG(LOG_DEBUG,
	    hash2hex(&hostId->hashPubKey, 
		     &hex));
#if DEBUG_CONNECTION
      LOG(LOG_DEBUG, 
	  "DEBUG: Marking host %s active.\n",
	  &hex);
#endif
      break;
    case STAT_UP:
      cronTime(&be->isAlive);
      break;
    default:
      LOG(LOG_WARNING, 
	  "WARNING: undefined status (%d)!\n",
	  be->status);
      break;
    }
  }
  MUTEX_UNLOCK(&lock);
}

/**
 * We received a sign of life from this host. 
 **/
void notifyPING(const HostIdentity * hostId) {
  BufferEntry * be;
  HexName hex;

  MUTEX_LOCK(&lock);
  be = lookForHost(hostId);
  if (be != NULL) {
    switch (be->status) {
    case STAT_DOWN:
      break;
    case STAT_WAITING_FOR_PONG:
      /* wrong message */
      break;
    case STAT_WAITING_FOR_PING:
      be->status = STAT_UP;
      CONNECTION_currentActiveHosts++;
      statChange(stat_number_of_connections,
		 1);
      cronTime(&be->isAlive);
      IFLOG(LOG_DEBUG,
	    hash2hex(&hostId->hashPubKey, 
		     &hex));
#if DEBUG_CONNECTION
      LOG(LOG_DEBUG, 
	  "DEBUG: Marking host %s active.\n",
	  &hex);
#endif
      break;
    case STAT_UP:
      cronTime(&be->isAlive);
      break;
    default:
      LOG(LOG_WARNING, 
	  "WARNING: unknown status!\n");
      break;
    }
  }
  MUTEX_UNLOCK(&lock);  
}

/**
 * Send a message to all directly connected nodes.
 *
 * @param message the message to send
 * @param priority how important is the message? The higher, the more important
 * @param maxdelay how long can we wait (max), in CRON-time (ms)
 **/
void broadcast(p2p_HEADER * message,
	       unsigned int priority,
	       unsigned int maxdelay) {
  unsigned int i;
  BufferEntry * tmp;

  MUTEX_LOCK(&lock);
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    tmp = CONNECTION_buffer_[i];
    while (tmp != NULL) {
      /* we need no sync here as we only read,
         and concurrent rw access does not hurt */
      if (tmp->status == STAT_UP) 
	unicast(&tmp->session.sender,
		message,
		priority,
		maxdelay);
      tmp = tmp->overflowChain;
    }
  }
  MUTEX_UNLOCK(&lock);    
}

/**
 * Send a message to a specific host (reply, enqueue)
 *
 * @param message the message to send (unencrypted!), first 2 bytes give size
 * @param hostId the identity of the receiver
 * @param priority how important is the message?
 * @param maxdelay how long can we wait (max), in CRON-time (ms)
 **/
void unicast(const HostIdentity * hostId,
	     p2p_HEADER * message,
	     unsigned int priority,
	     unsigned int maxdelay) {
  BufferEntry * be;
  HexName hex;

  if (hostIdentityEquals(&myIdentity, 
			 hostId)) {
    LOG(LOG_ERROR, 
	"ERROR: attempt to unicast to myself! (dropped)\n");
    breakpoint();
    return;
  }
  IFLOG(LOG_DEBUG,
	hash2hex(&hostId->hashPubKey, 
		 &hex));
#if DEBUG_CONNECTION
  LOG(LOG_DEBUG, 
      "DEBUG: unicasting to host %s message of type %d\n",
      &hex,
      ntohs(message->requestType));
#endif
  MUTEX_LOCK(&lock);
  be = lookForHost(hostId);
  if ( (be == NULL) || 
       (be->status == STAT_DOWN) ) {
    be = addHost(hostId, YES);   /* we *really* want to talk to this guy */
    if (be->status == STAT_DOWN)
      exchangeKey(be); /* note that exchangeKey
			  can fail and leave the buffer
			  in "down" state, e.g. if we don't
			  know any key for the other peer */
  }
  if (be->status != STAT_DOWN) {  
    be->prio += priority;
    appendToBuffer(be,
		   message,
		   maxdelay);
  }
  MUTEX_UNLOCK(&lock);
}

/**
 * Shutdown all connections (send HANGUPs, too).
 **/
void shutdownConnections() {
  MUTEX_LOCK(&lock);
  forAllConnectedHosts((BufferEntryCallback)&shutdownConnection, 
		       NULL);
  MUTEX_UNLOCK(&lock);
}

/**
 * Are we connected to this host?
 **/
int isConnected(const HostIdentity * hi) {
  BufferEntry * be;

  MUTEX_LOCK(&lock);
  be = lookForHost(hi);
  MUTEX_UNLOCK(&lock);
  if (be == NULL) {
    return NO;
  } else {
    return YES;
  }
}
	       
/**
 * Decipher data coming in from a foreign host.
 *
 * @param data the data to decrypt
 * @param size the size of the encrypted data
 * @param hostId the sender host that encrypted the data 
 * @param result where to store the decrypted data, must
 *        be at least of size data->len long
 * @returns the size of the decrypted data, SYSERR on error
 **/
int decryptFromHost(void * data,
		    const unsigned short size,
		    const HostIdentity * hostId,
		    void * result) {  
  BufferEntry * be;
  int res;
  HexName hex;

  statChange(stat_number_of_bytes_received, size);
  if ( (data == NULL) || 
       (hostId == NULL) ) {    
    LOG(LOG_FAILURE, 
	"CONNECTION: could not decrypt, message or hostId was NULL\n");
    breakpoint();
    return SYSERR;
  }
  IFLOG(LOG_DEBUG,
	hash2hex(&hostId->hashPubKey, 
		 &hex));
#if DEBUG_CONNECTION
  LOG(LOG_DEBUG, 
      "DEBUG: decrypting message from host %s\n",
      &hex);
#endif
  MUTEX_LOCK(&lock);
  be = lookForHost(hostId);
  if (be == NULL) {
    IFLOG(LOG_FAILURE,
	  hash2hex(&hostId->hashPubKey, 
		   &hex));
    LOG(LOG_FAILURE, 
	"FAILURE: decrypting message from host %s failed, no sessionkey!\n",
	&hex);
    MUTEX_UNLOCK(&lock);
    return SYSERR; /* could not decrypt */
  }
  res = decryptBlock(&be->skey, 
		     data,
		     size,
		     INITVALUE,
		     result);
  MUTEX_UNLOCK(&lock);
  return res;
}

/**
 * Compute the hashtable index of a host id.
 **/
unsigned int computeIndex(const HostIdentity * hostId) {
  unsigned int res = (((unsigned int)hostId->hashPubKey.a) & 
		      ((unsigned int)(CONNECTION_MAX_HOSTS_ - 1)));
  if (res >= CONNECTION_MAX_HOSTS_)
    errexit("FATAL: CONNECTION_MAX_HOSTS_ not power of 2? (%d)\n",
	    CONNECTION_MAX_HOSTS_);
  return res;
}

Mutex * getConnectionModuleLock() {
  return &lock;
}

/* end of connection.c */
